"""HTTP transport for sending telemetry to the collector."""

from __future__ import annotations

import json
import logging
import threading
from collections import deque
from typing import Any, Optional
from urllib.request import Request, urlopen
from urllib.error import URLError

logger = logging.getLogger("agentpulse")


class Transport:
    """Batched HTTP transport with background flushing.

    Accumulates events and flushes them in batches to minimize overhead.
    Uses only stdlib (urllib) to maintain zero-dependency constraint.
    """

    def __init__(
        self,
        endpoint: str,
        api_key: Optional[str] = None,
        flush_interval: float = 2.0,
        batch_size: int = 50,
    ) -> None:
        self._endpoint = endpoint.rstrip("/")
        self._api_key = api_key
        self._flush_interval = flush_interval
        self._batch_size = batch_size
        self._trace_queue: deque[dict[str, Any]] = deque()
        self._span_queue: deque[dict[str, Any]] = deque()
        self._lock = threading.Lock()
        self._timer: Optional[threading.Timer] = None
        self._closed = False
        self._start_flush_timer()

    def _start_flush_timer(self) -> None:
        if self._closed:
            return
        self._timer = threading.Timer(self._flush_interval, self._flush_loop)
        self._timer.daemon = True
        self._timer.start()

    def _flush_loop(self) -> None:
        self.flush()
        self._start_flush_timer()

    def send_trace(self, trace_data: dict[str, Any]) -> None:
        with self._lock:
            self._trace_queue.append(trace_data)
            if len(self._trace_queue) >= self._batch_size:
                self._flush_traces()

    def send_span(self, span_data: dict[str, Any]) -> None:
        with self._lock:
            self._span_queue.append(span_data)
            if len(self._span_queue) >= self._batch_size:
                self._flush_spans()

    def flush(self) -> None:
        with self._lock:
            self._flush_traces()
            self._flush_spans()

    def _flush_traces(self) -> None:
        if not self._trace_queue:
            return
        batch = list(self._trace_queue)
        self._trace_queue.clear()
        self._post(f"{self._endpoint}/v1/traces", batch)

    def _flush_spans(self) -> None:
        if not self._span_queue:
            return
        batch = list(self._span_queue)
        self._span_queue.clear()
        self._post(f"{self._endpoint}/v1/spans", batch)

    def _post(self, url: str, payload: list[dict[str, Any]]) -> None:
        headers = {"Content-Type": "application/json"}
        if self._api_key:
            headers["X-AgentPulse-Key"] = self._api_key

        data = json.dumps(payload).encode("utf-8")
        req = Request(url, data=data, headers=headers, method="POST")
        try:
            with urlopen(req, timeout=10) as resp:
                resp.read()
        except (URLError, OSError) as exc:
            logger.warning("AgentPulse: failed to send telemetry to %s: %s", url, exc)

    def close(self) -> None:
        self._closed = True
        if self._timer:
            self._timer.cancel()
        self.flush()



"""Data models for AgentPulse telemetry."""

from __future__ import annotations

import time
import uuid
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional


class SpanKind(str, Enum):
    LLM = "llm"
    TOOL = "tool"
    CUSTOM = "custom"


class TraceStatus(str, Enum):
    RUNNING = "running"
    SUCCESS = "success"
    ERROR = "error"


@dataclass
class Span:
    name: str
    kind: SpanKind
    trace_id: str
    id: str = field(default_factory=lambda: uuid.uuid4().hex)
    parent_span_id: Optional[str] = None
    started_at: float = field(default_factory=time.time)
    ended_at: Optional[float] = None
    input: Optional[Any] = None
    output: Optional[Any] = None
    model: Optional[str] = None
    tokens_in: Optional[int] = None
    tokens_out: Optional[int] = None
    cost_usd: Optional[float] = None
    error: Optional[str] = None

    def end(self, error: Optional[str] = None) -> None:
        self.ended_at = time.time()
        if error:
            self.error = error

    def set_output(self, output: Any) -> None:
        self.output = output

    def set_input(self, input_data: Any) -> None:
        self.input = input_data

    def to_dict(self) -> dict[str, Any]:
        kind_value = self.kind.value if isinstance(self.kind, SpanKind) else str(self.kind)
        return {
            "id": self.id,
            "trace_id": self.trace_id,
            "parent_span_id": self.parent_span_id,
            "name": self.name,
            "kind": kind_value,
            "started_at": self.started_at,
            "ended_at": self.ended_at,
            "input": self.input,
            "output": self.output,
            "model": self.model,
            "tokens_in": self.tokens_in,
            "tokens_out": self.tokens_out,
            "cost_usd": self.cost_usd,
            "error": self.error,
        }


@dataclass
class Trace:
    agent_name: Optional[str] = None
    id: str = field(default_factory=lambda: uuid.uuid4().hex)
    status: TraceStatus = TraceStatus.RUNNING
    started_at: float = field(default_factory=time.time)
    ended_at: Optional[float] = None
    total_tokens_in: int = 0
    total_tokens_out: int = 0
    total_cost_usd: float = 0.0
    metadata: Optional[dict[str, Any]] = None
    error: Optional[str] = None
    spans: list[Span] = field(default_factory=list)

    def end(self, status: TraceStatus = TraceStatus.SUCCESS, error: Optional[str] = None) -> None:
        self.ended_at = time.time()
        self.status = status
        if error:
            self.error = error
        for span in self.spans:
            self.total_tokens_in += span.tokens_in or 0
            self.total_tokens_out += span.tokens_out or 0
            self.total_cost_usd += span.cost_usd or 0.0

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "agent_name": self.agent_name,
            "status": self.status.value,
            "started_at": self.started_at,
            "ended_at": self.ended_at,
            "total_tokens_in": self.total_tokens_in,
            "total_tokens_out": self.total_tokens_out,
            "total_cost_usd": self.total_cost_usd,
            "metadata": self.metadata,
            "error": self.error,
        }


# Model pricing in USD per 1K tokens
MODEL_COSTS: dict[str, dict[str, float]] = {
    "gpt-4o": {"input": 0.0025, "output": 0.01},
    "gpt-4o-mini": {"input": 0.00015, "output": 0.0006},
    "gpt-4-turbo": {"input": 0.01, "output": 0.03},
    "gpt-4": {"input": 0.03, "output": 0.06},
    "gpt-3.5-turbo": {"input": 0.0005, "output": 0.0015},
    "claude-3-5-sonnet-20241022": {"input": 0.003, "output": 0.015},
    "claude-3-5-haiku-20241022": {"input": 0.0008, "output": 0.004},
    "claude-3-opus-20240229": {"input": 0.015, "output": 0.075},
    "claude-3-sonnet-20240229": {"input": 0.003, "output": 0.015},
    "claude-3-haiku-20240307": {"input": 0.00025, "output": 0.00125},
}


def calculate_cost(model: str, tokens_in: int, tokens_out: int) -> float:
    """Calculate cost for a model call. Returns 0.0 if model is unknown."""
    # Try exact match first, then prefix match
    costs = MODEL_COSTS.get(model)
    if not costs:
        for key, value in MODEL_COSTS.items():
            if model.startswith(key) or key.startswith(model):
                costs = value
                break
    if not costs:
        return 0.0
    return (tokens_in * costs["input"] + tokens_out * costs["output"]) / 1000



"""Trace context management using contextvars for async safety."""

from __future__ import annotations

from contextvars import ContextVar, Token
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .models import Span, Trace

_current_trace: ContextVar[Optional[Trace]] = ContextVar("agentpulse_trace", default=None)
_current_span: ContextVar[Optional[Span]] = ContextVar("agentpulse_span", default=None)


def get_current_trace() -> Optional[Trace]:
    return _current_trace.get()


def set_current_trace(trace: Optional[Trace]) -> Token[Optional[Trace]]:
    return _current_trace.set(trace)


def restore_trace(token: Token[Optional[Trace]]) -> None:
    _current_trace.reset(token)


def get_current_span() -> Optional[Span]:
    return _current_span.get()


def set_current_span(span: Optional[Span]) -> Token[Optional[Span]]:
    return _current_span.set(span)


def restore_span(token: Token[Optional[Span]]) -> None:
    _current_span.reset(token)



"""Decorators for tracing agent functions and tools."""

from __future__ import annotations

import asyncio
import functools
import logging
from typing import Any, Callable, Optional, TypeVar, overload

from .context import (
    get_current_span,
    get_current_trace,
    restore_span,
    restore_trace,
    set_current_span,
    set_current_trace,
)
from .models import Span, SpanKind, Trace, TraceStatus

logger = logging.getLogger("agentpulse")

F = TypeVar("F", bound=Callable[..., Any])


@overload
def trace(fn: F) -> F: ...


@overload
def trace(*, name: Optional[str] = None, metadata: Optional[dict[str, Any]] = None) -> Callable[[F], F]: ...


@overload
def trace(name: str) -> Callable[[F], F]: ...


def trace(
    fn: Optional[F] = None,
    *,
    name: Optional[str] = None,
    metadata: Optional[dict[str, Any]] = None,
) -> Any:
    """Decorator to trace an agent function.

    Creates a new trace (if none active) or a child span.

    Usage:
        @trace
        def my_agent(): ...

        @trace(name="research-agent")
        async def research(topic): ...
    """
    # Handle @trace("name") syntax — first positional arg is a string
    if fn is not None and isinstance(fn, str):
        return trace(name=fn)

    def decorator(func: F) -> F:
        trace_name = name or func.__name__

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            return await _run_traced(func, trace_name, metadata, args, kwargs, is_async=True)

        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            return _run_traced(func, trace_name, metadata, args, kwargs, is_async=False)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper  # type: ignore[return-value]
        return sync_wrapper  # type: ignore[return-value]

    if fn is not None and callable(fn):
        return decorator(fn)
    return decorator


def _run_traced(
    func: Callable[..., Any],
    trace_name: str,
    metadata: Optional[dict[str, Any]],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
    *,
    is_async: bool,
) -> Any:
    from .client import get_client

    client = get_client()
    existing_trace = get_current_trace()

    if existing_trace:
        # Nested: create a child span instead of a new trace
        span = Span(
            name=trace_name,
            kind=SpanKind.CUSTOM,
            trace_id=existing_trace.id,
            parent_span_id=(get_current_span() or Span(name="", kind=SpanKind.CUSTOM, trace_id="")).id
            if get_current_span()
            else None,
        )
        existing_trace.spans.append(span)
        span_token = set_current_span(span)

        if is_async:
            return _async_span_exec(func, span, span_token, args, kwargs)
        return _sync_span_exec(func, span, span_token, args, kwargs)

    # Top-level: create a new trace
    trace_obj = Trace(agent_name=trace_name, metadata=metadata)
    trace_token = set_current_trace(trace_obj)
    span_token = set_current_span(None)

    if is_async:
        return _async_trace_exec(func, trace_obj, trace_token, span_token, client, args, kwargs)
    return _sync_trace_exec(func, trace_obj, trace_token, span_token, client, args, kwargs)


def _sync_span_exec(
    func: Callable[..., Any],
    span: Span,
    span_token: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    try:
        result = func(*args, **kwargs)
        span.end()
        return result
    except Exception as exc:
        span.end(error=str(exc))
        raise
    finally:
        restore_span(span_token)


async def _async_span_exec(
    func: Callable[..., Any],
    span: Span,
    span_token: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    try:
        result = await func(*args, **kwargs)
        span.end()
        return result
    except Exception as exc:
        span.end(error=str(exc))
        raise
    finally:
        restore_span(span_token)


def _sync_trace_exec(
    func: Callable[..., Any],
    trace_obj: Trace,
    trace_token: Any,
    span_token: Any,
    client: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    try:
        result = func(*args, **kwargs)
        if client:
            client.end_trace(trace_obj, TraceStatus.SUCCESS)
        else:
            trace_obj.end(TraceStatus.SUCCESS)
        return result
    except Exception as exc:
        if client:
            client.end_trace(trace_obj, TraceStatus.ERROR, error=str(exc))
        else:
            trace_obj.end(TraceStatus.ERROR, error=str(exc))
        raise
    finally:
        restore_trace(trace_token)
        restore_span(span_token)


async def _async_trace_exec(
    func: Callable[..., Any],
    trace_obj: Trace,
    trace_token: Any,
    span_token: Any,
    client: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    try:
        result = await func(*args, **kwargs)
        if client:
            client.end_trace(trace_obj, TraceStatus.SUCCESS)
        else:
            trace_obj.end(TraceStatus.SUCCESS)
        return result
    except Exception as exc:
        if client:
            client.end_trace(trace_obj, TraceStatus.ERROR, error=str(exc))
        else:
            trace_obj.end(TraceStatus.ERROR, error=str(exc))
        raise
    finally:
        restore_trace(trace_token)
        restore_span(span_token)


@overload
def tool(fn: F) -> F: ...


@overload
def tool(*, name: Optional[str] = None) -> Callable[[F], F]: ...


@overload
def tool(name: str) -> Callable[[F], F]: ...


def tool(
    fn: Optional[F] = None,
    *,
    name: Optional[str] = None,
) -> Any:
    """Decorator to trace a tool function.

    Usage:
        @tool
        def search(query): ...

        @tool(name="web-search")
        async def search(query): ...
    """
    if fn is not None and isinstance(fn, str):
        return tool(name=fn)

    def decorator(func: F) -> F:
        tool_name = name or func.__name__

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            from .client import get_client
            client = get_client()
            if not client:
                return await func(*args, **kwargs)
            span = client.start_span(tool_name, SpanKind.TOOL)
            span_token = set_current_span(span)
            try:
                result = await func(*args, **kwargs)
                span.end()
                return result
            except Exception as exc:
                span.end(error=str(exc))
                raise
            finally:
                restore_span(span_token)

        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            from .client import get_client
            client = get_client()
            if not client:
                return func(*args, **kwargs)
            span = client.start_span(tool_name, SpanKind.TOOL)
            span_token = set_current_span(span)
            try:
                result = func(*args, **kwargs)
                span.end()
                return result
            except Exception as exc:
                span.end(error=str(exc))
                raise
            finally:
                restore_span(span_token)

        if asyncio.iscoroutinefunction(func):
            return async_wrapper  # type: ignore[return-value]
        return sync_wrapper  # type: ignore[return-value]

    if fn is not None and callable(fn):
        return decorator(fn)
    return decorator



"""Auto-patching for OpenAI client to track LLM calls."""

from __future__ import annotations

import asyncio
import functools
import logging
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from ..client import AgentPulse

from ..context import get_current_span, restore_span, set_current_span
from ..models import SpanKind, calculate_cost

logger = logging.getLogger("agentpulse")


def patch_openai(ap: AgentPulse, client: Optional[Any] = None) -> Optional[Any]:
    """Patch OpenAI client(s) for automatic span creation.

    If `client` is provided, wraps that specific instance and returns it.
    If `client` is None, monkey-patches the openai module globally.
    """
    try:
        import openai
    except ImportError:
        logger.warning("AgentPulse: openai package not installed, skipping patch")
        return client

    if client is not None:
        _patch_client_instance(ap, client)
        return client

    # Global patch: wrap the default clients
    _patch_module(ap, openai)
    return None


def _patch_client_instance(ap: AgentPulse, client: Any) -> None:
    """Patch a specific OpenAI client instance."""
    if hasattr(client, "chat") and hasattr(client.chat, "completions"):
        _wrap_completions(ap, client.chat.completions)
    if hasattr(client, "completions"):
        _wrap_completions(ap, client.completions)


def _patch_module(ap: AgentPulse, openai_module: Any) -> None:
    """Monkey-patch the openai module to wrap new client instances."""
    original_init = openai_module.OpenAI.__init__
    original_async_init = openai_module.AsyncOpenAI.__init__

    @functools.wraps(original_init)
    def patched_init(self: Any, *args: Any, **kwargs: Any) -> None:
        original_init(self, *args, **kwargs)
        _patch_client_instance(ap, self)

    @functools.wraps(original_async_init)
    def patched_async_init(self: Any, *args: Any, **kwargs: Any) -> None:
        original_async_init(self, *args, **kwargs)
        _patch_client_instance(ap, self)

    openai_module.OpenAI.__init__ = patched_init
    openai_module.AsyncOpenAI.__init__ = patched_async_init


def _wrap_completions(ap: AgentPulse, completions: Any) -> None:
    """Wrap the create method on a completions resource."""
    if not hasattr(completions, "create"):
        return

    original_create = completions.create
    is_async = asyncio.iscoroutinefunction(original_create)

    if is_async:
        @functools.wraps(original_create)
        async def traced_create(*args: Any, **kwargs: Any) -> Any:
            model = kwargs.get("model", "unknown")
            messages = kwargs.get("messages")
            span = ap.start_span(
                name=f"openai.{model}",
                kind=SpanKind.LLM,
                input_data=_safe_serialize_messages(messages),
            )
            span.model = model
            span_token = set_current_span(span)
            try:
                response = await original_create(*args, **kwargs)
                _extract_usage(span, response)
                return response
            except Exception as exc:
                span.end(error=str(exc))
                raise
            finally:
                restore_span(span_token)
    else:
        @functools.wraps(original_create)
        def traced_create(*args: Any, **kwargs: Any) -> Any:
            model = kwargs.get("model", "unknown")
            messages = kwargs.get("messages")
            span = ap.start_span(
                name=f"openai.{model}",
                kind=SpanKind.LLM,
                input_data=_safe_serialize_messages(messages),
            )
            span.model = model
            span_token = set_current_span(span)
            try:
                response = original_create(*args, **kwargs)
                _extract_usage(span, response)
                return response
            except Exception as exc:
                span.end(error=str(exc))
                raise
            finally:
                restore_span(span_token)

    completions.create = traced_create


def _extract_usage(span: Any, response: Any) -> None:
    """Extract token usage and cost from an OpenAI response."""
    usage = getattr(response, "usage", None)
    if usage:
        span.tokens_in = getattr(usage, "prompt_tokens", 0)
        span.tokens_out = getattr(usage, "completion_tokens", 0)
        if span.model:
            span.cost_usd = calculate_cost(span.model, span.tokens_in, span.tokens_out)

    # Extract output text
    choices = getattr(response, "choices", None)
    if choices and len(choices) > 0:
        message = getattr(choices[0], "message", None)
        if message:
            span.set_output(getattr(message, "content", None))

    span.end()


def _safe_serialize_messages(messages: Any) -> Any:
    """Safely serialize messages for storage, truncating large content."""
    if messages is None:
        return None
    if not isinstance(messages, list):
        return str(messages)[:1000]

    result = []
    for msg in messages:
        if isinstance(msg, dict):
            serialized = {**msg}
            if "content" in serialized and isinstance(serialized["content"], str):
                if len(serialized["content"]) > 500:
                    serialized["content"] = serialized["content"][:500] + "...[truncated]"
            result.append(serialized)
        else:
            result.append(str(msg)[:500])
    return result



"""AgentPulse - Lightweight observability for AI agents."""

from .client import AgentPulse, get_client
from .decorators import tool, trace
from .models import MODEL_COSTS, Span, SpanKind, Trace, TraceStatus, calculate_cost

__all__ = [
    "AgentPulse",
    "get_client",
    "trace",
    "tool",
    "Trace",
    "Span",
    "SpanKind",
    "TraceStatus",
    "calculate_cost",
    "MODEL_COSTS",
]

__version__ = "0.1.0"



"""AgentPulse client - the main entry point for the SDK."""

from __future__ import annotations

import atexit
import logging
from contextlib import contextmanager
from typing import Any, Generator, Optional

from .context import (
    get_current_span,
    get_current_trace,
    restore_span,
    set_current_span,
)
from .models import Span, SpanKind, Trace, TraceStatus
from .transport import Transport

logger = logging.getLogger("agentpulse")

_global_client: Optional[AgentPulse] = None


def get_client() -> Optional[AgentPulse]:
    return _global_client


class AgentPulse:
    """AgentPulse observability client.

    Usage:
        # Cloud
        ap = AgentPulse(api_key="ap_xxxxx")

        # Self-hosted
        ap = AgentPulse(endpoint="http://localhost:3000")
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        endpoint: str = "http://localhost:3000",
        flush_interval: float = 2.0,
        batch_size: int = 50,
        enabled: bool = True,
    ) -> None:
        global _global_client

        self.api_key = api_key
        self.endpoint = endpoint
        self.enabled = enabled
        self._transport: Optional[Transport] = None

        if enabled:
            self._transport = Transport(
                endpoint=endpoint,
                api_key=api_key,
                flush_interval=flush_interval,
                batch_size=batch_size,
            )
            atexit.register(self.shutdown)

        _global_client = self

    def start_trace(
        self,
        agent_name: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> Trace:
        trace = Trace(agent_name=agent_name, metadata=metadata)
        return trace

    def end_trace(
        self,
        trace: Trace,
        status: TraceStatus = TraceStatus.SUCCESS,
        error: Optional[str] = None,
    ) -> None:
        trace.end(status=status, error=error)
        if self._transport and self.enabled:
            self._transport.send_trace(trace.to_dict())
            for span in trace.spans:
                self._transport.send_span(span.to_dict())

    def start_span(
        self,
        name: str,
        kind: SpanKind = SpanKind.CUSTOM,
        input_data: Optional[Any] = None,
    ) -> Span:
        trace = get_current_trace()
        if not trace:
            logger.debug("AgentPulse: no active trace for span '%s'", name)
            # Create an orphan span with empty trace_id
            span = Span(name=name, kind=kind, trace_id="")
            if input_data is not None:
                span.set_input(input_data)
            return span

        parent = get_current_span()
        span = Span(
            name=name,
            kind=kind,
            trace_id=trace.id,
            parent_span_id=parent.id if parent else None,
        )
        if input_data is not None:
            span.set_input(input_data)
        trace.spans.append(span)
        return span

    @contextmanager
    def span(self, name: str, kind: SpanKind = SpanKind.CUSTOM) -> Generator[Span, None, None]:
        """Context manager for creating a custom span."""
        span = self.start_span(name, kind)
        token = set_current_span(span)
        try:
            yield span
        except Exception as exc:
            span.end(error=str(exc))
            raise
        else:
            span.end()
        finally:
            restore_span(token)

    @contextmanager
    def tool(self, name: str) -> Generator[Span, None, None]:
        """Context manager for tracking tool usage."""
        with self.span(name, kind=SpanKind.TOOL) as span:
            yield span

    def patch_openai(self, client: Optional[Any] = None) -> Optional[Any]:
        """Patch OpenAI client for automatic LLM call tracking.

        If a client instance is passed, patches and returns that instance.
        If no client is passed, patches the openai module globally.
        """
        from .patches.openai import patch_openai
        return patch_openai(self, client)

    def patch_anthropic(self, client: Optional[Any] = None) -> Optional[Any]:
        """Patch Anthropic client for automatic LLM call tracking."""
        from .patches.anthropic import patch_anthropic
        return patch_anthropic(self, client)

    def flush(self) -> None:
        if self._transport:
            self._transport.flush()

    def shutdown(self) -> None:
        if self._transport:
            self._transport.close()
